@@ -561,6 +561,7 @@ get_window_rect (NetkWindow * window,
     double width_ratio, height_ratio;
     int x, y, width, height;
     NetkWorkspace *workspace;
+    GdkRectangle unclipped_win_rect;
 
     workspace = netk_window_get_workspace (window);
     if (workspace == NULL)
@@ -593,10 +594,12 @@ get_window_rect (NetkWindow * window,
     if (height < 3)
         height = 3;
 
-    rect->x = x;
-    rect->y = y;
-    rect->width = width;
-    rect->height = height;
+    unclipped_win_rect.x = x;
+    unclipped_win_rect.y = y;
+    unclipped_win_rect.width = width;
+    unclipped_win_rect.height = height;
+
+    gdk_rectangle_intersect ((GdkRectangle *) workspace_rect, &unclipped_win_rect, rect);
 }
 
 static void
@@ -609,16 +612,18 @@ draw_window (GdkDrawable * drawable, GtkWidget * widget, NetkWindow * win,
 
     is_active = netk_window_is_active (win);
 
-    gdk_draw_rectangle (drawable,
-                        is_active ?
-                        widget->style->bg_gc[GTK_STATE_SELECTED] :
-                        widget->style->bg_gc[GTK_STATE_NORMAL],
-                        TRUE,
-                        winrect->x + 1, winrect->y + 1,
-                        winrect->width - 2, winrect->height - 2);
+    if ((winrect->width > 2) && (winrect->height > 2))
+    {
+        gdk_draw_rectangle (drawable,
+                            is_active ?
+                            widget->style->bg_gc[GTK_STATE_SELECTED] :
+                            widget->style->bg_gc[GTK_STATE_NORMAL],
+                            TRUE,
+                            winrect->x + 1, winrect->y + 1,
+                            winrect->width - 2, winrect->height - 2);
+    }
 
     icon = netk_window_get_icon (win);
-
     icon_w = icon_h = 0;
 
     if (icon)
@@ -691,13 +696,16 @@ draw_window (GdkDrawable * drawable, GtkWidget * widget, NetkWindow * win,
         }
     }
 
-    gdk_draw_rectangle (drawable,
-                        is_active ?
-                        widget->style->fg_gc[GTK_STATE_SELECTED] :
-                        widget->style->fg_gc[GTK_STATE_NORMAL],
-                        FALSE,
-                        winrect->x, winrect->y,
-                        winrect->width - 1, winrect->height - 1);
+    if ((winrect->width > 0) && (winrect->height > 0))
+    {
+        gdk_draw_rectangle (drawable,
+                            is_active ?
+                            widget->style->fg_gc[GTK_STATE_SELECTED] :
+                            widget->style->fg_gc[GTK_STATE_NORMAL],
+                            FALSE,
+                            winrect->x, winrect->y,
+                            winrect->width - 1, winrect->height - 1);
+    }
 }
 
 static int
@@ -1153,11 +1161,11 @@ netk_pager_button_release (GtkWidget * widget, GdkEventButton * event)
 
             if (space)
             {
-                netk_window_move_to_workspace (pager->priv->drag_window,
-                                               space);
-                if (space ==
-                    netk_screen_get_active_workspace (pager->priv->screen))
+                netk_window_move_to_workspace (pager->priv->drag_window, space);
+                if (space == netk_screen_get_active_workspace (pager->priv->screen))
+                {
                     netk_window_activate (pager->priv->drag_window);
+                }
             }
         }
 
@@ -1169,24 +1177,35 @@ netk_pager_button_release (GtkWidget * widget, GdkEventButton * event)
         i = workspace_at_point (pager, event->x, event->y, &viewport_x,
                                 &viewport_y);
 
-        if (i >= 0)
+        if (i >= 0 && (space = netk_screen_get_workspace (pager->priv->screen, i)))
         {
-            space = netk_screen_get_workspace (pager->priv->screen, i);
-
-            if (space)
-            {
+            int screen_width, screen_height;
+            
+            /* Don't switch the desktop if we're already there */               
+            if (space != netk_screen_get_active_workspace (pager->priv->screen))
                 netk_workspace_activate (space);
+                
+            /* EWMH only lets us move the viewport for the active workspace,
+             * but we just go ahead and hackily assume that the activate
+             * just above takes effect prior to moving the viewport
+             */
 
-                /* EWMH only lets us move the viewport for the active workspace,
-                 * but we just go ahead and hackily assume that the activate
-                 * just above takes effect prior to moving the viewport
-                 */
-                netk_screen_move_viewport (pager->priv->screen,
-                                           viewport_x, viewport_y);
+            /* Transform the pointer location to viewport origin, assuming
+             * that we want the nearest "regular" viewport containing the
+             * pointer.
+             */                                
+            screen_width  = netk_screen_get_width  (pager->priv->screen);
+            screen_height = netk_screen_get_height (pager->priv->screen);
+            viewport_x = (viewport_x / screen_width)  * screen_width;
+            viewport_y = (viewport_y / screen_height) * screen_height;
 
-                if (pager->priv->drag_window)
-                    netk_window_activate (pager->priv->drag_window);
-            }
+            if (netk_workspace_get_viewport_x (space) != viewport_x ||
+                netk_workspace_get_viewport_y (space) != viewport_y)
+              netk_screen_move_viewport (pager->priv->screen, viewport_x, viewport_y);
+
+            if ((space == netk_screen_get_active_workspace (pager->priv->screen))
+                 && (pager->priv->drag_window))
+              netk_window_activate (pager->priv->drag_window);
         }
 
         if (pager->priv->drag_window)
@@ -1607,8 +1626,7 @@ static void
 netk_pager_connect_window (NetkPager * pager, NetkWindow * window)
 {
     g_signal_connect_object (G_OBJECT (window), "name_changed",
-                             G_CALLBACK (window_name_changed_callback), pager,
-                             0);
+                             G_CALLBACK (window_name_changed_callback), pager, 0);
     g_signal_connect_object (G_OBJECT (window), "state_changed",
                              G_CALLBACK (window_state_changed_callback),
                              pager, 0);
@@ -1616,8 +1634,7 @@ netk_pager_connect_window (NetkPager * pager, NetkWindow * window)
                              G_CALLBACK (window_workspace_changed_callback),
                              pager, 0);
     g_signal_connect_object (G_OBJECT (window), "icon_changed",
-                             G_CALLBACK (window_icon_changed_callback), pager,
-                             0);
+                             G_CALLBACK (window_icon_changed_callback), pager, 0);
     g_signal_connect_object (G_OBJECT (window), "geometry_changed",
                              G_CALLBACK (window_geometry_changed_callback),
                              pager, 0);
