@@ -4,6 +4,7 @@
  * Copyright (C) 2001 Havoc Pennington
  * Copyright (C) 2003 Kim Woelders
  * Copyright (C) 2003 Red Hat, Inc.
+ * Includes urgency notification based on Matthew Eaton's patch for libwnck
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -33,6 +34,7 @@
 #endif
 
 #include "netk-window.h"
+#include "netk-class-group.h"
 #include "netk-xutils.h"
 #include "netk-private.h"
 #include "netk-enum-types.h"
@@ -43,7 +45,7 @@
 
 static GHashTable *window_hash = NULL;
 
-/* Keep 0-7 in sync with the numbers in the WindowState enum. Yeah I'm
+/* Keep 0-9 in sync with the numbers in the WindowState enum. Yeah I'm
  * a loser.
  */
 #define COMPRESS_STATE(window)                          \
@@ -55,19 +57,24 @@ static GHashTable *window_hash = NULL;
     ((window)->priv->skip_taskbar        << 5) |        \
     ((window)->priv->is_sticky           << 6) |        \
     ((window)->priv->is_hidden           << 7) |        \
-    ((window)->priv->is_fullscreen       << 8) )
+    ((window)->priv->is_fullscreen       << 8) |        \
+    ((window)->priv->is_urgent           << 9) )
 
 struct _NetkWindowPrivate
 {
     Window xwindow;
     NetkScreen *screen;
     NetkApplication *app;
+    NetkClassGroup *class_group;
     Window group_leader;
     Window transient_for;
     char *name;
     char *icon_name;
     char *session_id;
     char *session_id_utf8;
+    char *res_class;
+    char *res_name;
+
     int pid;
     int workspace;
 
@@ -97,6 +104,7 @@ struct _NetkWindowPrivate
     guint is_sticky:1;
     guint is_hidden:1;
     guint is_fullscreen:1;
+    guint is_urgent : 1;
 
     /* _NET_WM_STATE_HIDDEN doesn't map directly into an
      * externally-visible state (it determines the WM_STATE
@@ -120,6 +128,8 @@ struct _NetkWindowPrivate
     guint need_update_actions:1;
     guint need_update_wintype:1;
     guint need_update_transient_for:1;
+    guint need_update_wmclass : 1;
+    guint need_update_wmhints : 1;
 };
 
 enum
@@ -156,6 +166,7 @@ static void update_workspace (NetkWindow * window);
 static void update_actions (NetkWindow * window);
 static void update_wintype (NetkWindow * window);
 static void update_transient_for (NetkWindow * window);
+static void update_wmclass (NetkWindow *window);
 static void unqueue_update (NetkWindow * window);
 static void queue_update (NetkWindow * window);
 static void force_update_now (NetkWindow * window);
@@ -231,8 +242,7 @@ netk_window_class_init (NetkWindowClass * klass)
         g_signal_new ("workspace_changed", G_OBJECT_CLASS_TYPE (object_class),
                       G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (NetkWindowClass,
                                                           workspace_changed),
-                      NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
-                      0);
+                      NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
 
     signals[ICON_CHANGED] =
         g_signal_new ("icon_changed", G_OBJECT_CLASS_TYPE (object_class),
@@ -276,6 +286,8 @@ netk_window_finalize (GObject * object)
     g_free (window->priv->name);
     g_free (window->priv->icon_name);
     g_free (window->priv->session_id);
+    g_free (window->priv->res_class);
+    g_free (window->priv->res_name);
 
     if (window->priv->app)
         g_object_unref (G_OBJECT (window->priv->app));
@@ -364,6 +376,8 @@ p_netk_window_create (Window xwindow, NetkScreen * screen)
     window->priv->need_update_actions = TRUE;
     window->priv->need_update_wintype = TRUE;
     window->priv->need_update_transient_for = TRUE;
+    window->priv->need_update_wmclass = TRUE;
+    window->priv->need_update_wmhints = TRUE;
     force_update_now (window);
 
     return window;
@@ -450,17 +464,14 @@ netk_window_get_xid (NetkWindow * window)
     return window->priv->xwindow;
 }
 
-/**
- * netk_window_get_session_id:
- * @window: a #NetkWindow
- *
- * Gets the session ID for @window in Latin-1 encoding.
- * NOTE: this is invalid UTF-8. You can't display this
- * string in a GTK widget without converting to UTF-8.
- * See netk_window_get_session_id_utf8().
- *
- * Return value: session ID in Latin-1
- **/
+NetkClassGroup *
+netk_window_get_class_group (NetkWindow *window)
+{
+  g_return_val_if_fail (NETK_IS_WINDOW (window), NULL);
+
+  return window->priv->class_group;
+}
+
 const char *
 netk_window_get_session_id (NetkWindow * window)
 {
@@ -527,6 +538,22 @@ netk_window_get_window_type (NetkWindow * window)
     return window->priv->wintype;
 }
 
+const char*
+netk_window_get_resource_class (NetkWindow *window)
+{
+  g_return_val_if_fail (NETK_IS_WINDOW (window), NULL);
+
+  return window->priv->res_class;
+}
+
+const char*
+netk_window_get_resource_name (NetkWindow *window)
+{
+  g_return_val_if_fail (NETK_IS_WINDOW (window), NULL);
+
+  return window->priv->res_name;
+}
+
 /**
  * netk_window_is_minimized:
  * @window: a #NetkWindow
@@ -1014,10 +1041,14 @@ get_icons (NetkWindow * window)
         window->priv->need_emit_icon_changed = TRUE;
 
         if (window->priv->icon)
+        {
             g_object_unref (G_OBJECT (window->priv->icon));
+        }
 
         if (window->priv->mini_icon)
+        {
             g_object_unref (G_OBJECT (window->priv->mini_icon));
+        }
 
         window->priv->icon = icon;
         window->priv->mini_icon = mini_icon;
@@ -1034,10 +1065,9 @@ netk_window_get_icon (NetkWindow * window)
 
     get_icons (window);
     if (window->priv->need_emit_icon_changed)
-        queue_update (window);  /* not done in get_icons since we call that from
-                                 * the update
-                                 */
-
+    {
+        queue_update (window);
+    }
     return window->priv->icon;
 }
 
@@ -1048,10 +1078,9 @@ netk_window_get_mini_icon (NetkWindow * window)
 
     get_icons (window);
     if (window->priv->need_emit_icon_changed)
-        queue_update (window);  /* not done in get_icons since we call that from
-                                 * the update
-                                 */
-
+    {
+        queue_update (window);
+    }
     return window->priv->mini_icon;
 }
 
@@ -1248,6 +1277,20 @@ p_netk_window_set_application (NetkWindow * window, NetkApplication * app)
     window->priv->app = app;
 }
 
+void
+p_netk_window_set_class_group (NetkWindow     *window,
+                             NetkClassGroup *class_group)
+{
+    g_return_if_fail (NETK_IS_WINDOW (window));
+    g_return_if_fail (class_group == NULL || NETK_IS_CLASS_GROUP (class_group));
+
+    if (class_group)
+        g_object_ref (G_OBJECT (class_group));
+    if (window->priv->class_group)
+        g_object_unref (G_OBJECT (window->priv->class_group));
+    window->priv->class_group = class_group;
+}
+
 void
 p_netk_window_process_property_notify (NetkWindow * window, XEvent * xevent)
 {
@@ -1261,25 +1304,21 @@ p_netk_window_process_property_notify (NetkWindow * window, XEvent * xevent)
         window->priv->need_update_wm_state = TRUE;
         queue_update (window);
     }
-    else if (xevent->xproperty.atom == XA_WM_NAME
-             || xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_NAME")
-             || xevent->xproperty.atom ==
-             p_netk_atom_get ("_NET_WM_VISIBLE_NAME"))
+    else if ((xevent->xproperty.atom == XA_WM_NAME) ||
+             (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_NAME")) ||
+             (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_VISIBLE_NAME")))
     {
         window->priv->need_update_name = TRUE;
         queue_update (window);
     }
-    else if (xevent->xproperty.atom == XA_WM_ICON_NAME
-             || xevent->xproperty.atom ==
-             p_netk_atom_get ("_NET_WM_ICON_NAME")
-             || xevent->xproperty.atom ==
-             p_netk_atom_get ("_NET_WM_VISIBLE_ICON_NAME"))
+    else if ((xevent->xproperty.atom == XA_WM_ICON_NAME) ||
+             (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_ICON_NAME")) ||
+             (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_VISIBLE_ICON_NAME")))
     {
         window->priv->need_update_icon_name = TRUE;
         queue_update (window);
     }
-    else if (xevent->xproperty.atom ==
-             p_netk_atom_get ("_NET_WM_ALLOWED_ACTIONS"))
+    else if (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_ALLOWED_ACTIONS"))
     {
         window->priv->need_update_actions = TRUE;
         queue_update (window);
@@ -1289,8 +1328,7 @@ p_netk_window_process_property_notify (NetkWindow * window, XEvent * xevent)
         window->priv->need_update_workspace = TRUE;
         queue_update (window);
     }
-    else if (xevent->xproperty.atom ==
-             p_netk_atom_get ("_NET_WM_WINDOW_TYPE"))
+    else if (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_WINDOW_TYPE"))
     {
         window->priv->need_update_wintype = TRUE;
         queue_update (window);
@@ -1301,14 +1339,25 @@ p_netk_window_process_property_notify (NetkWindow * window, XEvent * xevent)
         window->priv->need_update_wintype = TRUE;
         queue_update (window);
     }
-    else if (xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_ICON")
-             || xevent->xproperty.atom == p_netk_atom_get ("KWM_WIN_ICON")
-             || xevent->xproperty.atom == p_netk_atom_get ("WM_HINTS"))
+    else if (xevent->xproperty.atom == XA_WM_CLASS)
+    {
+        window->priv->need_update_wmclass = TRUE;
+        queue_update (window);
+    }
+    else if ((xevent->xproperty.atom == p_netk_atom_get ("_NET_WM_ICON")) ||
+             (xevent->xproperty.atom == p_netk_atom_get ("KWM_WIN_ICON")))
     {
         p_netk_icon_cache_property_changed (window->priv->icon_cache,
                                             xevent->xproperty.atom);
         queue_update (window);
     }
+    else if (xevent->xproperty.atom == p_netk_atom_get ("WM_HINTS"))
+    {
+        p_netk_icon_cache_property_changed (window->priv->icon_cache,
+                                            xevent->xproperty.atom);
+        window->priv->need_update_wmhints = TRUE;
+        queue_update (window);
+    }
 }
 
 void
@@ -1419,7 +1468,6 @@ update_state (NetkWindow * window)
         case NETK_WINDOW_SPLASHSCREEN:
             window->priv->skip_taskbar = TRUE;
             break;
-
         case NETK_WINDOW_TOOLBAR:
         case NETK_WINDOW_MENU:
         case NETK_WINDOW_UTILITY:
@@ -1428,11 +1476,11 @@ update_state (NetkWindow * window)
             /* Skip taskbar if the window is transient
              * for some main application window
              */
-            if (window->priv->transient_for != None &&
-                !window->priv->transient_for_root)
+            if (window->priv->transient_for != None)
+            {
                 window->priv->skip_taskbar = TRUE;
+            }
             break;
-
         case NETK_WINDOW_NORMAL:
             break;
     }
@@ -1721,6 +1769,39 @@ update_transient_for (NetkWindow * window)
     }
 }
 
+static void
+update_wmclass (NetkWindow *window)
+{
+    if (!window->priv->need_update_wmclass)
+      return;
+
+    window->priv->need_update_wmclass = FALSE;
+
+    g_free (window->priv->res_class);
+    g_free (window->priv->res_name);
+
+    p_netk_get_wmclass (window->priv->xwindow,
+                        &window->priv->res_class,
+                        &window->priv->res_name);
+}
+
+static void
+update_wmhints (NetkWindow *window)
+{
+    if (!window->priv->need_update_wmhints)
+      return;
+
+    window->priv->need_update_wmhints = FALSE;
+    if (p_netk_is_urgent (window->priv->xwindow))
+    {
+        window->priv->is_urgent = TRUE;
+    }
+    else
+    {
+        window->priv->is_urgent = FALSE;
+    }
+}
+
 static void
 force_update_now (NetkWindow * window)
 {
@@ -1770,6 +1851,8 @@ force_update_now (NetkWindow * window)
     old_state = COMPRESS_STATE (window);
     old_actions = window->priv->actions;
 
+    update_wmclass (window);
+    update_wmhints (window);
     update_transient_for (window);      /* wintype needs this to be first */
     update_wintype (window);
     update_wm_state (window);
@@ -1778,14 +1861,23 @@ force_update_now (NetkWindow * window)
                                  */
     update_workspace (window);  /* emits signals */
     update_actions (window);
+    
 
     get_icons (window);
 
     new_state = COMPRESS_STATE (window);
-
     if (old_state != new_state)
-        emit_state_changed (window, old_state ^ new_state, new_state);
+    {
+        NetkWindowState changed_state = old_state ^ new_state;
 
+        emit_state_changed (window, changed_state, new_state);
+        /* Hack, pwack, arck, crack...!!! */
+        if ((changed_state & NETK_WINDOW_STATE_SKIP_TASKLIST) &&
+            (old_state & NETK_WINDOW_STATE_SKIP_TASKLIST))
+        {
+            emit_geometry_changed (window);
+        }
+    }
     if (old_actions != window->priv->actions)
         emit_actions_changed (window, old_actions ^ window->priv->actions,
                               window->priv->actions);
